from typing import Callable

import torch
from torch.utils.data import DataLoader


def warmup(
    forward_call: Callable, data_loader: DataLoader, device: torch.device
) -> None:
    """
    Warms up the forward call for a few iterations.

    Inputs:
    - forward_call (Callable): The forward call we will be benchmarking.
    - data_loader (DataLoader): The data loader we use for the warmup. Should be the same as
        for the benchmarking.
    - device (torch.device): the device we are targetting.

    Outputs:
    None
    """
    counter = 0
    with torch.no_grad():
        for data, _ in data_loader:
            data = data.to(device)
            _ = forward_call(data)
            counter += 1
            if counter > 10:
                return



import argparse
import logging
from pathlib import Path
from typing import Tuple, Union

import torch

from ..conversions.select import MODEL_CONVERSION_OPTIONS
from ..utils.ipdb_hook import ipdb_sys_excepthook

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


# Define a custom argument type for a list of strings
def list_of_strings(arg):
    return arg.split(",")


def parse_benchmark_args() -> Tuple[argparse.Namespace, torch.device]:

    # Create a string represenation of the model conversion options
    # to add to the argparser description.
    string_rep_of_conv_options: str = "; \n".join(
        [f"{key}: {value}" for key, value in MODEL_CONVERSION_OPTIONS.items()]
    )
    valid_conversion_options = list(MODEL_CONVERSION_OPTIONS.keys()) + list(
        MODEL_CONVERSION_OPTIONS.values()
    )

    parser = argparse.ArgumentParser(description="Benchmark PyTorch Models")
    parser.add_argument(
        "--model-path",
        type=str,
        default=None,
        help="Path to the model file (e.g., mnist_cnn_quantized.pt, mnist_cnn_scripted.pt, mnist_cnn.pt)",
    )
    parser.add_argument(
        "--data-dir",
        type=str,
        default=None,
        help="Path to the directory containing samples for inference",
    )
    parser.add_argument(
        "--batch-size",
        type=int,
        default=64,
        metavar="N",
        help="input batch size for benchmarking (default: 30)",
    )
    parser.add_argument(
        "--n-samples",
        type=int,
        default=2048,
        help="Total number of samples to process (default: 300)",
    )
    parser.add_argument(
        "--no-cuda",
        action="store_true",
        default=False,
        help="disables CUDA acceleration",
    )
    parser.add_argument(
        "--no-mps",
        action="store_true",
        default=False,
        help="disables MPS acceleration",
    )
    parser.add_argument(
        "--conversions",
        type=list_of_strings,
        # choices=[str(i) for i in MODEL_CONVERSION_OPTIONS.keys()] + list(MODEL_CONVERSION_OPTIONS.values()),
        default=None,
        help=f"""The model option you would like to benchmark. These are integers that correspond 
to different transforms, or their string names. MUltiple options can be selected, e.g. --conversions
0,2,EAGER . The mapping is this:\n{string_rep_of_conv_options}""",
    )
    parser.add_argument(
        "--ipdb",
        action="store_true",
        default=False,
        help="Enable the ipdb system exception hook",
    )

    args = parser.parse_args()

    if args.ipdb:
        # Add an ipdb hook to the sys.excepthook, which will throw one into an ipdb shell when an
        # exception is raised.
        ipdb_sys_excepthook()

    if args.model_path is not None:
        args.model_path = Path(args.model_path)

    use_cuda = not args.no_cuda and torch.cuda.is_available()
    use_mps = not args.no_mps and torch.backends.mps.is_available()

    # If no conversion options are provided, we use all available options
    if not args.conversions:
        conversions = valid_conversion_options
    else:
        conversions = args.conversions

    # Checks on the model conversion options
    assert isinstance(
        conversions, (list, int, str)
    ), "Please select a valid option for the model conversion"
    if not isinstance(conversions, list):
        conversions = [conversions]

    error_msg = (
        lambda conversion: f"Please select a valid option for the model conversion, {conversion} not in {valid_conversion_options}. Call `-h` for help."
    )
    # Convert all selected conversion options to a list of strings. I.e., all ints become strings
    # We also check that the provided conversion options are valid
    selected_conversions = []
    for conversion in conversions:
        if isinstance(conversion, str) and conversion.isnumeric():
            conversion = int(conversion)
            assert conversion in valid_conversion_options, error_msg(conversion)
            selected_conversions.append(MODEL_CONVERSION_OPTIONS[conversion])
        elif isinstance(conversion, str) and not conversion.isnumeric():
            assert conversion in valid_conversion_options, error_msg(conversion)
            selected_conversions.append(conversion)
        elif isinstance(conversion, int):
            assert conversion in valid_conversion_options, error_msg(conversion)
            selected_conversions.append(MODEL_CONVERSION_OPTIONS[conversion])
        else:
            raise ValueError(error_msg(conversion))

    # Convert the list of selected conversions to a "pretty" string for logging
    format_list = lambda lst: ", ".join(lst[:-1]) + (
        " and " + lst[-1] if len(lst) > 1 else lst[0] if lst else ""
    )
    logger.info(
        f"{format_list(selected_conversions)} model conversions selected for benchmarking\n"
    )

    args.conversions = selected_conversions

    if use_cuda:
        device = torch.device("cuda")
    elif use_mps:
        device = torch.device("mps")
    else:
        device = torch.device("cpu")

    return args, device



import logging
from typing import Callable, List

import torch
import torch.fx as fx
import torch.quantization as tq
from torch.ao.quantization.qconfig_mapping import QConfigMapping
from torch.ao.quantization.quantize_fx import prepare_qat_fx

from .utils.checks.type import check_model_type
from .utils.qconfigs import fake_quant_act, fixed_0255, learnable_act, learnable_weights

# One needs to set their quantization backend engine to what is appropriate for their system.
# torch.backends.quantized.engine = 'x86'
torch.backends.quantized.engine = "qnnpack"

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def get_fake_quantized_model_forward_call(
    model: torch.nn.Module,
    data: torch.Tensor,
) -> Callable:
    """
    Quantize the model using FX Graph mode, and return the forward call.

    Inputs:
    - model (torch.nn.Module): The model to quantize.
    - data (torch.Tensor): Sample data to feed through the model for tracing and PTQ.

    Returns:
    - forward (Callable): the forward call of the FX Graph quantized model.
    """

    # FX graph mode fake-quantization
    fx_model: fx.GraphModule = fx_quantize(model, data)

    # Return forward call
    return fx_model.forward


def fx_quantize(
    model: torch.nn.Module,
    data: torch.Tensor,
) -> Callable:
    """
    Wrapper around the FX Graph mode quantization of the model, which will deal with error
    handling.

    Returns a fx-graph traced quantized model.

    Inputs:
    - model (torch.nn.Module): The model to quantize.
    - data (torch.Tensor): A sample of input data to use for tracing.

    Outputs:
    - fx_model (torch.fx.GraphModule): The quantized model in FX graph mode.
    """
    # FX graph mode fake-quantization
    try:
        fx_model: fx.GraphModule = _fx_quantize(model, data)
    except Exception as e:
        error_msg = f"""Fake quantization failed. Check that your model is traceable, and make sure 
that your quantization backend engine is set to match your hardware. E.g. 
torch.backends.quantized.engine = "qnnpack" | "x86". Error: {e}"""
        logger.error(error_msg)

    # Check model type
    check_model_type(fx_model, fx.GraphModule)

    # Make sure to send the model to the desired device
    fx_model = fx_model.to(data.device)
    return fx_model


def _fx_quantize(
    model: torch.nn.Module,
    data: torch.Tensor,
) -> torch.fx.GraphModule:
    """
    Quantizes the model with FX graph tracing. Does not do PTQ or QAT, and
    just provides default quantization parameters that are entirely not optimized.

    Returns a fx-graph traced quantized model.

    Inputs:
    - model (torch.nn.Module): The model to quantize.
    - data (torch.Tensor): A sample of input data to use for tracing.

    Outputs:
    - fx_model (torch.fx.GraphModule): The quantized model in FX graph mode.

    """

    logger.info("Quantizing model with FX graph mode")

    # These values define the quantization scales of the activations and weights. They are set
    # to arbitrary values for the sake of this benchmarking.
    default_act_scale = 2.0
    default_weight_scale = 2.0

    # Define global (default) qconfig
    qconfig_global = tq.QConfig(
        activation=learnable_act(range=default_act_scale),
        weight=tq.default_fused_per_channel_wt_fake_quant,
    )

    # Assign global qconfig
    qconfig_mapping = QConfigMapping().set_global(qconfig_global)

    # We loop through the modules so that we can access the `out_channels`/`out_features` attribute
    for name, module in model.named_modules():
        # Convolutional layers
        if hasattr(module, "out_channels"):
            qconfig = tq.QConfig(
                activation=learnable_act(range=default_act_scale),
                weight=learnable_weights(
                    range=default_weight_scale, channels=module.out_channels
                ),
            )
            qconfig_mapping.set_module_name(name, qconfig)
        # Linear layers
        elif hasattr(module, "out_features"):
            qconfig = tq.QConfig(
                activation=learnable_act(range=default_act_scale),
                weight=learnable_weights(
                    range=default_weight_scale, channels=module.out_features
                ),
            )
            qconfig_mapping.set_module_name(name, qconfig)

    # Do symbolic tracing and quantization
    model.eval()
    fx_model = prepare_qat_fx(model, qconfig_mapping, (data,))

    # Prints the graph as a table
    if logger.root.level <= logging.DEBUG:
        logger.debug("\nFake quantized model graph:\n")
        logger.debug(fx_model.graph.print_tabular())
    return fx_model



from typing import Any


def check_model_type(model: Any, expected_type: Any):
    """
    Checks that the model is of the expected type.

    Inputs:
    - model (Any): the model whose type to check
    - expected_type (Any): the class we expect

    Outputs:
    None
    """

    assert isinstance(
        model, expected_type
    ), f"model must be of type {expected_type}, got {type(model)}"



import logging
import time

import torch
import torch.quantization as tq
from tqdm import tqdm

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def PTQ(
    model: torch.nn.Module,
    device: torch.device,
    dataloader: torch.utils.data.DataLoader,
):
    """
    Perform PTQ on the model using the given dataloader.
    This works in place on the model.

    Inputs:
    - model (torch.nn.Module): The model to quantize.
    - device (torch.device): The device to run the model on.
    - dataloader (DataLoader): The DataLoader to get a sample of data from.
    """

    model.eval()
    # Ensure fake quantization enabled
    model.apply(tq.enable_fake_quant)

    # Ensure model is on the correct device
    model.to(device)

    # Enable PTQ observers
    for module in model.modules():
        if hasattr(module, "observer_enabled") or hasattr(module, "static_enabled"):
            module.enable_observer()

    start_time = time.time()
    with torch.no_grad():
        for data, target in tqdm(dataloader, desc="PTQ"):
            data, target = data.to(device), target.to(device)
            _ = model(data)
    end_time = time.time()
    logging.info(f"PTQ time: {end_time - start_time:.2f}s")

    # Disable PTQ observers
    for module in model.modules():
        if hasattr(module, "observer_enabled") or hasattr(module, "static_enabled"):
            module.disable_observer()



import logging
import traceback
from typing import Any, Dict, List, Union

import torch
from torch.utils.data import DataLoader

from .benchmark import benchmark
from .conversions.select import MODEL_CONVERSION_OPTIONS
from .dataloader.create import create_single_tensor_dataloader
from .utils.checks import check_consistent_batch_size, check_inputs
from .utils.times import inference_time_benchmarking  # should we use this?

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def benchmark_model(
    model: torch.nn.Module,
    config: Dict[str, Any],
    conversions: Union[List[str], None] = None,
    data: Union[torch.Tensor, None] = None,
    data_loader: Union[DataLoader, None] = None,
) -> Dict[str, Dict[str, Any]]:
    """
    Benchmark the model on different conversion methods. If provided, the dataloader will be used.
    Else, a random dataloader will be created, in which case the `data` tensor must be provided
    to initialize the dataloader.
    The model is benchmarked on the given conversion methods, and the results are returned.
    The `config` dict must contain the following:
    - n_samples: int     # The number of samples to benchmark on
    - batch_size: int    # The batch size to use for benchmarking

    Inputs:
    - model (torch.nn.Module): The model to benchmark.
    - config (Dict[str, Any]): The configuration for the benchmarking. This contains the number of
        samples to benchmark on, and the batch size to use for benchmarking.
    - conversions (List[str]): The list of conversion methods to benchmark. If None, all of the
        available conversion methods will be benchmarked.
    - data (torch.Tensor): The data to use for benchmarking. If provided, and the data loader has
            not been provided, then the data shape will be used as the basis for the data loader.
    - data_loader (DataLoader): The DataLoader to get samples of data from. If provided, this will
            be used. Else, a random dataloader will be created.

    Outputs:
    - all_results (Dict[str, Dict[str, Any]]): The results of the benchmarking for each conversion method.
        The key is the conversion method, and the value is a tuple containing the total elapsed
        time, the total time taken, the total number of samples, and the throughput of the model.
        If the conversion method failed, the value will be a dictionary containing the error and
        traceback.
    """
    # Set to eval mode
    model.eval()

    # We determine the device to run the model on
    # NOTE: this will only work for single-device set ups. Benchmarking on multiple devices is not
    # currently supported.
    device: torch.device = next(model.parameters()).device

    # Check the inputs
    check_inputs(model, config, conversions, data, data_loader)

    # If the conversions are not provided, we use all available conversions
    if conversions is None:
        conversions = list(MODEL_CONVERSION_OPTIONS.values())

    # The number of samples to benchmark on, batch size, and whether or not to give verbose logging
    n_samples: int = config["n_samples"]
    batch_size: int = config["batch_size"]

    # Creates a dataloader with random data, of the same size as the input data sample
    # If the data_loader has been provided by the user, we use that one
    if not data_loader:
        data_loader = create_single_tensor_dataloader(
            tensor_size=data.size(),
            num_tensors=n_samples,
            random_type="normal",
            random_params={"mean": 0.0, "std": 2.0},
            batch_size=config["batch_size"],
        )

    all_results: Dict[str, Dict[str, float]] = {}
    for conversion_method in conversions:
        check_consistent_batch_size(conversion_method, n_samples, batch_size)

        logger.info(f"Benchmarking model using conversion: {conversion_method}")
        try:
            result: Dict[str, float] = benchmark(
                model, conversion_method, device, data_loader, n_samples
            )
            result["status"] = "success"
            all_results[conversion_method] = result
        except Exception as e:
            # If there is an error, we log the error. In the returned "results", we include the
            # full traceback
            error_msg = (
                f"Benchmarking conversion {conversion_method} failed. Error: {e}"
            )
            logger.error(error_msg)
            all_results[conversion_method] = {
                "status": "error",
                "error": e,
                "traceback": traceback.format_exc(),
            }

    return all_results



import logging
import tempfile
from pathlib import Path
from typing import Any, Callable

import torch

from .options.compile import (
    get_compiled_forward_call_eager_fallback,
    get_compiled_model_forward_call,
)
from .options.export_aotinductor import (
    get_export_aot_inductor_forward_call,
    get_quant_export_aot_inductor_forward_call,
)
from .options.export_compile import (
    get_export_compiled_forward_call,
    get_export_compiled_forward_call_eager_fallback,
)
from .options.export_eager import get_export_eager_forward_call
from .options.export_quant import get_quant_exported_forward_call
from .options.fake_quant import get_fake_quantized_model_forward_call
from .options.onnx import get_onnx_dynamo_forward_call, get_onnx_forward_call
from .options.quant_convert import get_converted_quantized_model_forward_call
from .options.tensorrt import get_tensorrt_dynamo_forward_call
from .options.utils.checks.imports import (
    check_onnxrt,
    check_openxla,
    check_tensort,
    check_tvm,
)
from .options.jit_trace import get_jit_traced_model_forward_call
from .options.torchscript import get_torch_scripted_model_forward_call


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

MODEL_CONVERSION_OPTIONS = {
    0: "EAGER",
    1: "EXPORT+EAGER",
    2: "ONNX_CPU",
    3: "ONNX_GPU",
    4: "ONNX+DYNAMO_EXPORT",
    5: "COMPILE_CUDAGRAPH",
    6: "COMPILE_INDUCTOR_DEFAULT",
    7: "COMPILE_INDUCTOR_REDUCE_OVERHEAD",
    8: "COMPILE_INDUCTOR_MAX_AUTOTUNE",
    9: "COMPILE_INDUCTOR_EAGER_FALLBACK",
    10: "COMPILE_ONNXRT",
    11: "COMPILE_OPENXLA",
    12: "COMPILE_TVM",
    13: "EXPORT+AI8WI8_FLOAT_QUANTIZED",
    14: "EXPORT+AI8WI8_FLOAT_QUANTIZED+AOT_INDUCTOR",
    15: "EXPORT+AI8WI8_FLOAT_QUANTIZED+RUN_DECOMPOSITION",
    16: "EXPORT+AI8WI8_FLOAT_QUANTIZED+RUN_DECOMPOSITION+AOT_INDUCTOR",
    17: "EXPORT+AI8WI8_STATIC_QUANTIZED",
    18: "EXPORT+AI8WI8_STATIC_QUANTIZED+AOT_INDUCTOR",
    19: "EXPORT+AI8WI8_STATIC_QUANTIZED+RUN_DECOMPOSITION",
    20: "EXPORT+AI8WI8_STATIC_QUANTIZED+RUN_DECOMPOSITION+AOT_INDUCTOR",
    21: "EXPORT+AOT_INDUCTOR",
    22: "EXPORT+COMPILE_CUDAGRAPH",
    23: "EXPORT+COMPILE_INDUCTOR_DEFAULT",
    24: "EXPORT+COMPILE_INDUCTOR_REDUCE_OVERHEAD",
    25: "EXPORT+COMPILE_INDUCTOR_MAX_AUTOTUNE",
    26: "EXPORT+COMPILE_INDUCTOR_DEFAULT_EAGER_FALLBACK",
    27: "EXPORT+COMPILE_ONNXRT",
    28: "EXPORT+COMPILE_OPENXLA",
    29: "EXPORT+COMPILE_TVM",
    30: "NATIVE_CONVERT_AI8WI8_STATIC_QUANTIZED",
    31: "NATIVE_FAKE_QUANTIZED_AI8WI8_STATIC",
    32: "TENSORRT",
    33: "JIT_TRACE",
    34: "TORCH_SCRIPT",
}


def select_forward_call_function(
    model: Any,
    conversion: str,
    data: torch.Tensor,
    device: torch.device,
) -> Callable:
    """
    Get the forward call function for the model, given a conversion type.

    Inputs:
    - model (Any): The model to get the forward call for.
    - conversion (str): The conversion method to use for the model.
    - data (torch.Tensor): A sample of data to pass through the model, which may be needed for
    some of the export methods.
    - device (torch.device): The device to run the benchmarking on. This is sometimes required
    for a conversion method.

    Outputs:
    - forward (Callable): The forward call function for the model.
    """
    match conversion:
        ###############
        # WITH EXPORT #
        ###############
        case "EXPORT+COMPILE_INDUCTOR_DEFAULT":
            forward = get_export_compiled_forward_call(
                model, data, backend="inductor-default"
            )

        case "EXPORT+COMPILE_INDUCTOR_REDUCE_OVERHEAD":
            forward = get_export_compiled_forward_call(
                model, data, backend="inductor-reduce-overhead"
            )

        case "EXPORT+COMPILE_INDUCTOR_MAX_AUTOTUNE":
            forward = get_export_compiled_forward_call(
                model, data, backend="inductor-max-autotune"
            )

        case "EXPORT+COMPILE_CUDAGRAPH":
            forward = get_export_compiled_forward_call(
                model, data, backend="cudagraphs"
            )

        case "EXPORT+COMPILE_ONNXRT":
            # Very much an evolving API, not guaranteed to work.
            check_onnxrt()
            forward = get_export_compiled_forward_call(model, data, backend="onnxrt")

        case "EXPORT+COMPILE_OPENXLA":
            # Check if torch-xla is installed
            check_openxla()
            forward = get_export_compiled_forward_call(model, data, backend="openxla")

        case "EXPORT+COMPILE_TVM":
            check_tvm()
            forward = get_export_compiled_forward_call(model, data, backend="tvm")

        case "EXPORT+COMPILE_INDUCTOR_DEFAULT_EAGER_FALLBACK":
            forward = get_export_compiled_forward_call_eager_fallback(
                model,
                data,
                backend="inductor-default",
            )

        case "EXPORT+AOT_INDUCTOR":
            forward = get_export_aot_inductor_forward_call(model, data, device)

        case "EXPORT+EAGER":
            forward = get_export_eager_forward_call(model, data)

        case "ONNX+DYNAMO_EXPORT":
            forward = get_onnx_dynamo_forward_call(model, data)

        case "EXPORT+AI8WI8_STATIC_QUANTIZED":
            forward = get_quant_exported_forward_call(
                model, data, int_or_dequant_op="int", run_decompositions=False
            )

        case "EXPORT+AI8WI8_FLOAT_QUANTIZED":
            forward = get_quant_exported_forward_call(
                model, data, int_or_dequant_op="dequant", run_decompositions=False
            )

        case "EXPORT+AI8WI8_STATIC_QUANTIZED+AOT_INDUCTOR":
            forward = get_quant_export_aot_inductor_forward_call(
                model, data, device, int_or_dequant_op="int", run_decompositions=False
            )

        case "EXPORT+AI8WI8_FLOAT_QUANTIZED+AOT_INDUCTOR":
            forward = get_quant_export_aot_inductor_forward_call(
                model,
                data,
                device,
                int_or_dequant_op="dequant",
                run_decompositions=False,
            )

        case "EXPORT+AI8WI8_STATIC_QUANTIZED+RUN_DECOMPOSITION":
            # The difference with training (i.e. inference=False) is that at the end we re-run
            # torch.export and then run `run_decompositions`, with the hope it might shorted the graph
            # a bit.
            forward = get_quant_exported_forward_call(
                model, data, int_or_dequant_op="int", run_decompositions=True
            )

        case "EXPORT+AI8WI8_FLOAT_QUANTIZED+RUN_DECOMPOSITION":
            forward = get_quant_exported_forward_call(
                model, data, int_or_dequant_op="dequant", run_decompositions=True
            )

        case "EXPORT+AI8WI8_STATIC_QUANTIZED+RUN_DECOMPOSITION+AOT_INDUCTOR":
            forward = get_quant_export_aot_inductor_forward_call(
                model, data, device, int_or_dequant_op="int", run_decompositions=True
            )

        case "EXPORT+AI8WI8_FLOAT_QUANTIZED+RUN_DECOMPOSITION+AOT_INDUCTOR":
            forward = get_quant_export_aot_inductor_forward_call(
                model,
                data,
                device,
                int_or_dequant_op="dequant",
                run_decompositions=True,
            )

        ##################
        # WITHOUT EXPORT #
        ##################
        case "COMPILE_INDUCTOR_DEFAULT":
            forward = get_compiled_model_forward_call(
                model, data, backend="inductor-default"
            )

        case "COMPILE_INDUCTOR_REDUCE_OVERHEAD":
            forward = get_compiled_model_forward_call(
                model, data, backend="inductor-reduce-overhead"
            )

        case "COMPILE_INDUCTOR_MAX_AUTOTUNE":
            forward = get_compiled_model_forward_call(
                model, data, backend="inductor-max-autotune"
            )

        case "COMPILE_CUDAGRAPH":
            forward = get_compiled_model_forward_call(model, data, backend="cudagraphs")

        case "COMPILE_ONNXRT":
            # Very much an evolving API, not guaranteed to work.
            check_onnxrt()
            forward = get_compiled_model_forward_call(model, data, backend="onnxrt")

        case "COMPILE_OPENXLA":
            check_openxla()
            forward = get_compiled_model_forward_call(model, data, backend="openxla")

        case "COMPILE_TVM":
            check_tvm()
            forward = get_compiled_model_forward_call(model, data, backend="tvm")

        case "COMPILE_INDUCTOR_DEFAULT_EAGER_FALLBACK":
            forward = get_compiled_forward_call_eager_fallback(
                model, data, backend="inductor-default"
            )

        case "EAGER":
            # Regular eager model forward call
            forward = model.forward

        case "TENSORRT":
            check_tensort()
            forward = get_tensorrt_dynamo_forward_call(model, data)

        case "ONNX_CPU":
            # We create temporary file to save the onnx model
            with tempfile.TemporaryDirectory() as tmpdirname:
                onnx_model_path = Path(f"{tmpdirname}/model.onnx")
                onnx_backend = "CPUExecutionProvider"
                forward = get_onnx_forward_call(
                    model, data, onnx_model_path, onnx_backend
                )

        case "ONNX_GPU":
            # We create temporary file to save the onnx model
            with tempfile.TemporaryDirectory() as tmpdirname:
                onnx_model_path = Path(f"{tmpdirname}/model.onnx")
                onnx_backend = "CUDAExecutionProvider"
                forward = get_onnx_forward_call(
                    model, data, onnx_model_path, onnx_backend
                )

        case "NATIVE_CONVERT_AI8WI8_STATIC_QUANTIZED":
            if device.type != "cpu":
                logger.warning(
                    "PyTorch native quantized model conversion is only supported for CPUs currently"
                )
            forward = get_converted_quantized_model_forward_call(model, data)

        case "NATIVE_FAKE_QUANTIZED_AI8WI8_STATIC":
            forward = get_fake_quantized_model_forward_call(model, data)

        case "JIT_TRACE":
            forward = get_jit_traced_model_forward_call(model, data)

        case "TORCH_SCRIPT":
            forward = get_torch_scripted_model_forward_call(model)

        case _:
            error_msg = f"The option {conversion} is not supported"
            assert conversion in MODEL_CONVERSION_OPTIONS.values(), error_msg
            raise NotImplementedError(error_msg)

    return forward



import sys
import traceback

import ipdb


def ipdb_sys_excepthook():
    """
    When called this function will set up the system exception hook.
    This hook throws one into an ipdb breakpoint if and where a system
    exception occurs in one's run.

    E.g.
    >>> ipdb_sys_excepthook()
    """

    def info(type, value, tb):
        """
        System excepthook that includes an ipdb breakpoint.
        """
        if hasattr(sys, "ps1") or not sys.stderr.isatty():
            # we are in interactive mode or we don't have a tty-like
            # device, so we call the default hook
            sys.__excepthook__(type, value, tb)
        else:
            # we are NOT in interactive mode, print the exception...
            traceback.print_exception(type, value, tb)
            print
            # ...then start the debugger in post-mortem mode.
            # pdb.pm() # deprecated
            ipdb.post_mortem(tb)  # more "modern"

    sys.excepthook = info



import logging
from typing import Optional, Tuple

import torch
import torch.fx as fx

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def save_fake_quantized_model(model: torch.nn.Module, path: str) -> None:
    """
    Saves the state dictionary of a quantized model to a specified path.

    Args:
        model (torch.nn.Module): The quantized model to be saved.
        path (str): The path where the model state dict will be saved.
    """
    model.eval()
    torch.save(model.state_dict(), path)
    logging.info(f"Quantized model saved to {path}")


def replace_node_with_target(
    fx_model: fx.GraphModule, target_node_name: str, target_module: torch.nn.Module
) -> None:
    """
    Replaces a node in the FX graph with a target module. For example, this can be useful for
    replacing quantization modules with custom modules.

    Args:
        fx_model (fx.GraphModule): The FX graph module containing the node to replace.
        target_node_name (str): The name of the target node to be replaced.
        target_module (torch.nn.Module): The new module to replace the target node.

    Raises:
        ValueError: If the node with the specified target name is not found.
    """

    def find_target_node() -> Optional[fx.Node]:
        """Find the node with the specified target name."""
        for node in fx_model.graph.nodes:
            if node.op == "call_module" and node.target == target_node_name:
                return node
        return None

    modules = dict(fx_model.named_modules())
    target_fx_node = find_target_node()

    if target_fx_node:
        replace_node_module(modules, target_fx_node, target_module)
        logging.info(f"Replaced node {target_node_name} with new module.")
    else:
        error_message = f"Could not find node with target {target_node_name}"
        logging.error(error_message)
        raise ValueError(error_message)


def replace_node_module(
    modules: dict, node: fx.Node, new_module: torch.nn.Module
) -> None:
    """
    Replaces a module in the given module dictionary with a new module.

    Args:
        modules (dict): A dictionary of modules.
        node (fx.Node): The node representing the target module to be replaced.
        new_module (torch.nn.Module): The new module to replace the target module.
    """
    assert isinstance(node.target, str), "Node target must be a string"
    parent_name, name = _parent_name(node.target)
    setattr(modules[parent_name], name, new_module)
    logging.info(f"Replaced module {node.target} in parent {parent_name}")


def _parent_name(target: str) -> Tuple[str, str]:
    """
    Splits a qualname into parent path and last atom.

    Args:
        target (str): The qualified name to split.

    Returns:
        Tuple[str, str]: The parent path and the last atom of the target.
    """
    *parent, name = target.rsplit(".", 1)
    return parent[0] if parent else "", name


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    # Example model, path, and target node for demonstration
    class ExampleModel(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.conv = torch.nn.Conv2d(1, 1, 3)

        def forward(self, x):
            return self.conv(x)

    example_model = ExampleModel()
    new_conv = torch.nn.Conv2d(1, 1, 3)
    fx_model = fx.symbolic_trace(example_model)

    # Replace the node in the fx model
    try:
        replace_node_with_target(fx_model, "conv", new_conv)
        logging.info("Node replacement completed successfully.")
    except ValueError as e:
        logging.error(f"Node replacement failed: {e}")



import logging
import time
from typing import Any, Dict

import torch
import torch._dynamo
from torch.utils.data import DataLoader
from tqdm import tqdm

from ..conversions.select import select_forward_call_function
from ..utils.data import get_sample_data
from ..utils.times import inference_time_benchmarking  # should we use this?
from .log import log_results
from .warmup import warmup

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def benchmark(
    model: torch.nn.Module,
    conversion: str,
    device: torch.device,
    data_loader: DataLoader,
    n_samples: int,
) -> Dict[str, float]:
    """
    Benchmark the model using the given data loader. This function will benchmark the model using the
    given conversion method.

    Inputs:
    - model (torch.nn.Module): The model to benchmark.
    - conversion (str): The conversion method to use for benchmarking.
    - device (torch.device): The device we are targetting.
    - data_loader (DataLoader): The DataLoader to get samples of data from.
    - n_samples (int): The number of samples to benchmark on.

    Outputs:
    - total_elapsed_time (float): The total elapsed time for the benchmark.
    - total_time (float): The total time taken for inference.
    - total_samples (int): The total number of samples benchmarked.
    - throughput (float): The throughput of the model.
    """
    total_inf_time = 0.0
    total_samples = 0
    num_batches = 0

    # Get sample of data from dataloader
    data = get_sample_data(data_loader, device)

    # Get the forward call of the model, which we will benchmark. We also return the device we will
    # benchmark on, since some conversions are only supported for certain devices, e.g.
    # PyTorch native quantized conversions requires CPU
    forward_call = select_forward_call_function(model, conversion, data, device)

    logger.info(f"Benchmarking {conversion} on {device}")

    # Clear all caches, etc.
    torch._dynamo.reset()

    # Warmup
    warmup(forward_call, data_loader, device)

    # Benchmarking loop
    start_time = time.perf_counter()  # Start timing for the entire process
    with torch.no_grad():
        for data, _ in tqdm(data_loader, desc=f"Benchmarking {conversion} on {device}"):
            if total_samples >= n_samples:
                break

            # data = data.to(device, non_blocking=True)
            data = data.to(device)
            batch_start_time = time.perf_counter()
            _ = forward_call(data)
            # if conversion in ["EXPORT+INFERENCE+AI8WI8_FLOAT_QUANTIZED", "EXPORT+TRAINING+AI8WI8_FLOAT_QUANTIZED"]:
            #     import ipdb; ipdb.set_trace()
            batch_end_time = time.perf_counter()

            batch_size = min(data.size(0), n_samples - total_samples)
            total_inf_time += batch_end_time - batch_start_time
            total_samples += batch_size
            num_batches += 1

            if total_samples >= n_samples:
                break

    end_time = time.perf_counter()  # End timing for the entire process

    total_elapsed_time = end_time - start_time
    throughput = total_samples / total_inf_time if total_elapsed_time > 0 else 0
    result: Dict[str, Any] = {
        "device": device,
        "total_elapsed_time": total_elapsed_time,
        "total_inf_time": total_inf_time,
        "total_samples": total_samples,
        "batch_size": data.shape[0],
        "throughput": throughput,
    }
    if logger.root.level <= logging.DEBUG:
        log_results(result)

    return result



import logging

import torch
from torch.export.exported_program import ExportedProgram

from ....utils.setup_logging import suppress_output

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def get_exported_model(model, data: torch.Tensor) -> ExportedProgram:
    """
    Export the model using torch.export.

    Inputs:
    - model (torch.nn.Module): The model to export
    - data (torch.Tensor): A sample of data to feed through the model for tracing

    Outputs:
    model (torch.export.Model): The exported model
    """

    logger.info("Running torch.export on the model")

    # Call torch export, which decomposes the forward pass of the model
    # into a graph of Aten primitive operators
    with suppress_output(logger.root.level >= logging.DEBUG):
        model = torch.export.export(model, (data,))

    logger.debug("Model graph:")
    if logger.root.level <= logging.DEBUG:
        logger.debug(model.graph.print_tabular())

    return model



from typing import Any, Callable

import torch
import torch.fx as fx
import torch_tensorrt


def get_tensorrt_dynamo_forward_call(model: Any, data: torch.Tensor) -> Callable:
    """
    Get the forward call function for the model using TensorRT.

    Inputs:
    - model (Any): The model to get the forward call for.
    - data (torch.Tensor): A sample of data to pass through the model.

    Outputs:
    - forward (Callable): The forward call function for the model.
    """
    # Compile the model using TensorRT
    trt_gm: fx.GraphModule = torch_tensorrt.dynamo.compile(model, data)

    return trt_gm.forward



from typing import List, Union

import torch
from torch.utils.data import DataLoader


def check_inputs(
    model: torch.nn.Module,
    config: dict,
    conversions: Union[List[str], None],
    data: Union[torch.Tensor, None],
    data_loader: Union[DataLoader, None],
):
    """
    Check the inputs to the benchmark_model function.

    Inputs:
    - model (torch.nn.Module): The model to benchmark.
    - config (dict): The configuration for the benchmarking.
    - conversions (list): The list of conversion methods to benchmark.
    - data (torch.Tensor): The data to use for benchmarking (initialising the dataloader).
    - data_loader (DataLoader): The DataLoader to get samples of data from.

    Outputs:
    None
    """
    assert isinstance(model, torch.nn.Module), "The model must be a torch.nn.Module"
    assert isinstance(config, dict), "The config must be a dictionary"
    assert isinstance(conversions, (list, type(None))), "The conversions must be a list"
    assert isinstance(
        data, (torch.Tensor, type(None))
    ), "The data must be a torch.Tensor"
    assert isinstance(
        data_loader, (DataLoader, type(None))
    ), "The data_loader must be a DataLoader"

    # Check the configuration
    assert "batch_size" in config, "The batch size must be provided in the config"
    assert (
        "n_samples" in config
    ), "The number of samples (n_samples) to benchmark on must be provided in the config"

    # Either the `data` Tensor must be provided, or a data loader
    if data is None:
        assert (
            data_loader is not None
        ), "If data is not provided, the data_loader must be provided"
    if data_loader is None:
        assert (
            data is not None
        ), "If data_loader is not provided, a data tensor must be provided"
    else:
        assert (
            data is None
        ), "If a data loader is provided, the data tensor must be None"


def check_consistent_batch_size(
    conversion_method: str, n_samples: int, batch_size: int
) -> None:
    """
    Check that the batch size will always be consistent, which is a requirement for certain
    conversion methods.

    Inputs:
    - conversion_method (str): the conversion method being evaluated.
    - n_samples (int): the number of samples to use during benchmarking. This may have to be a perfect
        multiple of the batch size, depending on the conversion method.
    - batch_size (int): the batch size to use during benchmarking.
    """
    if "COMPILE" in conversion_method:
        if n_samples % batch_size != 0:
            error_msg = f"""n_samples must be a multiple of batch_size for compilation based methods, 
otherwise the last batch will fail due to having a different shape. n-samples was {n_samples}, and 
batch size was {batch_size} """
            raise ValueError(error_msg)



import logging
from contextlib import contextmanager
from typing import Callable, Dict, Literal, Union

import torch
import torch._dynamo
import torch.fx as fx
from torch.export.exported_program import ExportedProgram

from ...utils.setup_logging import suppress_output
from .utils.checks.type import check_model_type

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def get_compiled_model(
    model: Union[torch.nn.Module, fx.GraphModule, ExportedProgram],
    data: torch.Tensor,
    backend: Literal[str],
) -> Callable:
    """
    Compile the model using torch.compile.

    Inputs:
    - model (torch.nn.Module): The model to export
    - data (torch.Tensor): A sample of data to feed through the model
    - backend (Literal['cudagraphs', 'inductor', 'onnxrt', 'openxla', 'tvm']): the backend for
        torch.compile to use.

    Outputs:
    model (torch._dynamo.eval_frame.OptimizedModule): The compiled model
    """
    logger.info(f"Running torch.compile [{backend} backend] on the model")
    check_model_type(model, (torch.nn.Module, fx.GraphModule, ExportedProgram))

    torch._dynamo.reset()

    # Set the compilation settings
    compile_settings: Dict[str, str] = get_compile_settings(backend)

    # Compile the model, with suppressed internal logs if logging is above Debug level.
    with suppress_output(logger.root.level >= logging.DEBUG):
        with torch.no_grad():
            model = torch.compile(model, **compile_settings)

            # Feed some data through the model to make sure it works
            _ = model(data)

    # Print model graph
    if logger.root.level <= logging.DEBUG:
        logger.debug("Model graph:")
        logger.debug(model.graph.print_tabular())

    check_model_type(model, torch._dynamo.eval_frame.OptimizedModule)

    return model


def get_compiled_model_forward_call(
    model: Union[torch.nn.Module, fx.GraphModule, ExportedProgram],
    data: torch.Tensor,
    backend: Literal[str],
) -> Callable:
    """
    Run torch.compile in the model, and get its forward call.

    Inputs:
    - model (torch.nn.Module): The model to export
    - data (torch.Tensor): A sample of data to feed through the model
    - backend (Literal['cudagraphs', 'inductor', 'onnxrt', 'openxla', 'tvm']): the backend for
        torch.compile to use.

    Outputs:
    forward (Callable): the forward call of the compiled model.
    """

    model = get_compiled_model(model, data, backend)
    return model.forward


def get_compiled_forward_call_eager_fallback(
    model: Union[torch.nn.Module, fx.GraphModule, ExportedProgram],
    data: torch.Tensor,
    backend: Literal[str],
) -> Callable:
    """
    Run torch.compile in the model, and get its forward call. If dynamo errors occur, we fallback
    on eager mode by wrapping it with a context manager.

    Inputs:
    - model (torch.nn.Module): The model to export
    - data (torch.Tensor): A sample of data to feed through the model
    - backend (Literal['cudagraphs', 'inductor', 'onnxrt', 'openxla', 'tvm']): the backend for
        torch.compile to use.

    Outputs:
    forward (Callable): the forward call of the compiled model.
    """
    with suppress_dynamo_errors():
        return get_compiled_model_forward_call(model, data, backend)


def get_compile_settings(backend: Literal[str] = "inductor-default") -> Dict[str, str]:
    """
    Get the compilation settings for each torch.dynamo backend choice.

    Inputs:
    - backend (Literal[str]): The backend to use for torch.compile. Currently supported options in
        PyTorch are given by torch._dynamo.list_backends():
        ['cudagraphs', 'inductor', 'onnxrt', 'openxla', 'tvm']
        However, we have also split 'inductor' into 'inductor-default', 'inductor-max-autotune', and
        'inductor-reduce-overhead'. See here for an explanation of each:
        https://pytorch.org/get-started/pytorch-2.0/#user-experience

    Outputs:
    - compile_settings (Dict[str, str]): The returned compilation settings.
    """
    match backend:
        case "inductor-default":
            compile_settings = {
                "mode": "default",
                "backend": "inductor",
            }
        case "inductor-reduce-overhead":
            compile_settings = {
                "mode": "reduce-overhead",
                "fullgraph": True,
                "backend": "inductor",
            }

        case "inductor-max-autotune":
            import torch._inductor.config

            torch._inductor.config.max_autotune_gemm_backends = "ATEN,TRITON,CPP"
            torch._inductor.config.max_autotune = True
            # TRITON backend fails with conv2d layers during max-autotune.
            torch._inductor.config.max_autotune_conv_backends = "ATEN"
            compile_settings = {
                "mode": "max-autotune",
                "fullgraph": True,
                "backend": "inductor",
                # Options can also be fed in via a dict, see here for options: https://github.com/pytorch/pytorch/blob/main/torch/_inductor/config.py
                # However, I found this to be slower than using the "preset" options, so I think feeding
                # in any options will prevent any tuning from happening. This is why the settings
                # are controlled via global settings above.
                # options: {},
            }

        case "cudagraphs" | "openxla" | "tvm":
            compile_settings = {
                "fullgraph": True,
                "backend": backend,
            }
        case "onnxrt":
            # TODO: debug why CUDA does not work on this.
            # See here for the accepted options for ONNXRT backend:
            # https://github.com/pytorch/pytorch/blob/05c1f37188b1923ee3e48634cb1aaa3e976e2d0f/torch/onnx/_internal/onnxruntime.py#L689
            from torch.onnx._internal.onnxruntime import OrtBackendOptions

            # Configure the ORT backend options. See here for options:
            # https://github.com/pytorch/pytorch/blob/main/torch/_inductor/config.py
            ort_options = OrtBackendOptions(
                # Specify preferred execution providers in priority order
                preferred_execution_providers=[
                    # ("CUDAExecutionProvider", {"device_id": 0}),  # For GPU execution
                    "CPUExecutionProvider",  # Fallback to CPU
                ],
                # Automatically infer execution providers from input/output devices
                infer_execution_providers=True,
                # Optional: Set to True if you want to pre-allocate output memory
                preallocate_output=True,
                # Disable AOT autograd for training support
                use_aot_autograd=False,
            )
            compile_settings = {
                "fullgraph": True,
                "backend": backend,
                "options": ort_options,
            }
        case _:
            raise ValueError(f"{backend} is not a valid option")

    return compile_settings


@contextmanager
def suppress_dynamo_errors():
    """
    Context manager to temporarily suppress torch._dynamo errors.
    This will have the execution fall back to eager mode.
    """
    # Store the original setting
    original_setting = torch._dynamo.config.suppress_errors
    try:
        # Set suppress_errors to True
        torch._dynamo.config.suppress_errors = True
        yield
    finally:
        # Restore the original setting
        torch._dynamo.config.suppress_errors = original_setting
